;----------------------------------------------------------------------------
; MBus Master I2C routines for TI-82 & CrASH 1.6
; Version 0.99.1 (public beta) 18.11.1999
; Copyright (C) 1999 Vasantha Crabb <_merlin@operamail.com>
; Copyright (C) 1998 Osma Suominen <ozone@iki.fi>
;----------------------------------------------------------------------------
;
;  This library is free software; you can redistribute it and/or
;  modify it under the terms of the GNU Library General Public
;  License as published by the Free Software Foundation; either
;  version 2 of the License, or (at your option) any later version.
;
;  This library is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;  Library General Public License for more details.
;
;  You should have received a copy of the GNU Library General Public
;  License along with this library; if not, write to the Free
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;----------------------------------------------------------------------------
;
; MI2C : Main I2C Multimaster master routine to read/write data at HL
;
; Usage: When calling, the registers are used as follows:
;        A : Protocol id ($00 or $01 for Raw Write and Read mode)
;        B : Data length (Write mode)
;    or BC : 16-bit data length (Raw Write and Read mode) (B=high C=low)
;        D : Destination (slave) address
;        E : 2nd byte of slave address when using 10-bit extended addressing
;       HL : Pointer to memory location where to start data read/write
;
;        Also put the calculator's own address at (I2CAddress).
;        This will be sent with every packet, except in the raw data modes.
;        Use $0 as the address if your calculator doesn't have one, i.e.
;        you only use the master routines.
;
;        The variable (memory location) I2CSpeed determines the speed at
;        which individual data bits are sent. $01 is fastest, $FF is second to
;        slowest and $00 is slowest (it's a DJNZ loop). A speed setting of
;        1 currently gives a data rate of around 2,2kBytes/sec with an
;        unturboed TI-85. A setting of 0 drops the data rate to
;        about 100 Bytes/sec (but you should never need that slow a rate).
;
;        Variables:
;        I2CAddress             Holds this calculator's address
;                               Not needed if ONLY_RAW is defined
;
;        I2CSpeed               Determines the transmission speed
;                               Not needed if ZERO_WAIT_STATE is defined
;
; The data type depends on the desired mode. The mode can be:
; - Write mode with software protocol (used when accessing other calculators)
; - Raw write mode: sending raw data to e.g. I2C chips  (addr: xxxxxxx0)
; - Raw read mode: reading raw data from e.g. I2C chips (addr: xxxxxxx1)
;
; When an error (occupied bus, no acknowledge or loss of arbitration)
; occurs a retry is made. After two retries the routines give up.
; This error situation is indicated by the Carry flag being 1.
;
; The master routines can be customized with C-style #define's, which are
; supported at least by TASM. The following #define's are available:
; (README.TXT contains a more detailed description of each.)
;
; #define ONLY_RAW              ; disable the protocol mode routines *
; #define NO_RAW                ; disable the raw mode routines
; #define ZERO_WAIT_STATE       ; disable all delays                 *
; #define NO_SYNCH              ; disable clock synchronization      *
; #define 10_BIT_ADDR           ; enable 10-bit extended addressing support
;
; The #define's marked with an asterisk (*) should only be used with the
; raw mode, to increase speed and make the routines smaller.
;
; Notes on "pinging" I2C devices:
; To test for a device at a given address with a packet that has no
; data, you _must_ use the Write mode. Set the destination address
; to the address you want to test, the data length to $00 and the protocol
; id to something else than $00, e.g. $FF. Then CALL_(MI2C). If CY=0 the
; device acknowledged, if CY=1 it didn't. As usual, a failure leads to
; two retries to improve reliability.
;
; This feature can be used to probe for a free I2C bus address or to list all
; the devices on the bus. It sends a one-byte packet (only the address) so
; it can be used for both other TI's and I2C chips, even though Write mode
; is used for it. This is a special use of the Write mode.
;
; NOTE!
; Ping packets should _never_ be sent to read addresses (xxxxxxx1), because
; the address byte causes the addressed slave to start sending data. If the
; MSB of the data is a 0, the master can't create a STOP condition, and the
; bus gets locked.
;
; Notes on Write mode:
; The routines will automatically add a one-byte checksum to the data and
; the receive routines at the other end will immediately check this checksum.
; If the checksums don't match, the last (checksum) data byte is NACKed,
; leading to an error situation and a retry at the sending (master) end.
; Because of the memory usage at the receiving end, the maximum size for a
; packet is 128 bytes (not including checksum, which is never stored in
; memory), limiting the maximum freeform data size to 125 bytes.
; Data length byte values > 125 might be used for special purposes in
; the future. One such purpose is splitting a packet of data into several,
; smaller packets if the link is too unreliable to handle large packets
; without errors. This can happen e.g. with an IR link. This has to be dealt
; with later.
;
; In the raw modes it is possible to write/read up to 64KBytes in a single
; packet. The data length is stored in BC. A length of zero is interpreted
; as 64KBytes.
;
; To write/read a single packet, use protocol id $00. If you want to send
; another packet without a STOP in between, with a repeated START condition,
; use $01 for the first packets and $00 for the last packet. This is very
; useful if you want to access e.g. a memory chip. You first send it the
; address, and then write or read the data without sending a STOP in between.
; If you sent a STOP after sending the address, another master could take
; over the bus, change the address and then free the bus. The address would
; now be different from the one you expected.
;
; Warning! Be extremely careful when reading raw data! If you set the read
; length incorrectly, you may overwrite some precious data! If you set the
; data length to 0, you will end up overwriting the whole memory (except
; that at some point the routine will overwrite itself and therefore stop,
; most probably hanging the calculator)!
; (Of course, always keep a backup...)
;
; The routines do not destroy the data in any of the registers.
; (Except for the flags, as the Carry flag is used for output.)

#define OldD ix-1
#define OldE ix-2
#define OldA ix-3

; Main routine

MI2C:
    push    ix
    ld      ix,0
    add     ix,sp                   ; IX <- SP
    push    de
    ld      e,3                     ; Number of tries
MI2CRetry:
    push    af
    push    bc
    push    hl
    push    de                      ; D is the only one that's conserved

#ifndef ONLY_RAW
#ifndef NO_RAW
    and     $FE
    jr      nz,MI2CWrite            ; Prot. id != 0,1 => Write w/protocol
#endif
#endif

#ifndef NO_RAW
    call    MI2CStart

; Support for Extended 10-bit addressing - currently untested
#ifdef 10_BIT_ADDR
    ld      a,$0F                   ; 00001111
    or      d
    inc     a                       ; $FF becomes $00
    jr      nz,MI2C7BitAddr
    ld      a,d
    and     $F6
    ld      d,a
    call    MI2CSend                ; Send 1st byte of dest addr (bit 0 = 0)
    ld      d,e
    call    MI2CSend                ; Send 2nd byte
    ld      a,(OldD)
    rra
    jr      nc,MI2CSWLoop           ; Write mode - continue as normal
    call    MI2CStart               ; Repeated Start condition
    ld      d,(OldD)
#endif
#endif

#ifndef NO_RAW
MI2C7BitAddr:
    call    MI2CSend
    rr      d
    jr      c,MI2CSRLoop

MI2CSWLoop:
    ld      d,(hl)
    inc     hl
    call    MI2CSend
    dec     bc
    ld      a,b
    or      c
    jr      nz,MI2CSWLoop
    jr      MI2CSuccess

MI2CSRLoop:
    call    MI2CRecv
    ld      (hl),d
    inc     hl
    dec     bc
    ld      a,b
    or      c
    jr      nz,MI2CSRLoop
    jr      MI2CSuccess

#endif

#ifndef ONLY_RAW
MI2CWrite:
    ld      e,0                     ; Init checksum counter
    call    MI2CStart

    call    MI2CSend                ; Send slave address (already in D)

    xor     a                       ; Check for zero-length packets
    cp      b
    jr      z,MI2CSuccess           ; Don't continue if length is zero

    cp      d                       ; A=0, CY set if D>0
    ccf                             ; CY set if D=0 (general call)
    ld      a,(I2CAddress)
    adc     a,0                     ; Add 1 to own address if it's a general call

    ld      d,a                     ; Send own address
    call    MI2CSend

    ld      d,b                     ; Send data length byte
    inc     d                       ; Length byte includes protocol id byte
    call    MI2CSend
    ld      d,(OldA)
    call    MI2CSend                ; Send protocol id byte

MI2CWriteLoop:
    ld      d,(hl)
    call    MI2CSend                ; Send data byte(s) (count=b)
    inc     hl
    djnz    MI2CWriteLoop
    ld      d,e
    call    MI2CSend                ; Send checksum byte
#endif

MI2CSuccess:                        ; STOP and return to calling application

#ifndef NO_RAW
    ld      a,(OldA)
    dec     a                       ; Is the protocol id =1?
    jr      z,MI2CNoStop            ; If it is, don't send a STOP
#endif
    call    MI2CStop                ; STOP - transfer OK
MI2CNoStop:
    pop     de

    pop     hl
    pop     bc
    pop     af
    pop     de
    pop     ix
    or      a                       ; CF=0: OK
    ret

; Reads D
; Destroys AF
; Determines ACK/NACK by looking at BC

#ifndef NO_RAW
MI2CRecv:
    push    bc

    ld      a,c                     ; Is it the last byte to be read?
    dec     a
    or      b
    cp      1                       ; NACK (CY=1) if it's the last byte

    ld      c,4                     ; 00000100, read mode flags

    ld      d,$FF
    jr      MI2CByte
#endif

; Sends D
; Counts the checksum in E (not required in raw mode but doesn't harm)
; Destroys AF

MI2CSend:
    push    bc
    ld      c,0                     ; Sets write mode flags

#ifndef ONLY_RAW
    ld      a,e                     ; Count the checksum
    xor     d
    ld      e,a
#endif

    scf                             ; Must be done for a valid Ack

; Subroutine used by MI2CSend and MI2CRecv
; Sends/receives byte in D according to C: 4=read 0=write
; In read operations D must be set to $FF, and CY
; determines if it's an ACK (0) or a NACK (1)
; In write operations CY must be set (1)
; Returns only after a valid Ack, in error situations jumps to
; error handlers.
; Output: D = sent/received byte
; Traps: AF destroyed, BC popped one extra time

MI2CByte:
    ld      b,9
#ifndef ZERO_WAIT_STATE
    push    hl
    ld      hl,I2CSpeed
#endif
MI2CBit:
    ld      a,$E8                   ; SCL=0 SDA=1
    rl      d                       ; Next bit to/from D
    jr      c,MI2CBit1
    or      $14                     ; SCL=0 SDA=0

#ifdef NO_SYNCH
MI2CBit1:
    out     (P_LINK),a              ; Data available before clock pulse
    push    bc
    ld      b,a                     ; Save A in B (push would save F also)
    and     $D4                     ; SCL=1
    out     (P_LINK),a              ; Clock pulse starts

    in      a,(P_LINK)
    or      c
    and     5                       ; Set Z on ack/error
    rra                             ; CY=sent/rcvd bit

#else
MI2CBit1:
    out     (P_LINK),a              ; Data available before clock pulse
    and     $D4                     ; SCL=1
    out     (P_LINK),a              ; Clock pulse starts
    or      $E8
    push    bc
    ld      b,a                     ; Save A in B (push would save F also)

MI2CBitLoop:
    in      a,(P_LINK)              ; Waitfor SCL=1 to synchronize speed w/ slave
    rra
    rra
    jr      nc,MI2CBitLoop
    or      c
    rlca                            ; Set Z on ack/error, CY=sent/rcvd byte

#endif

    ld      a,b                     ; Restore A

#ifndef ZERO_WAIT_STATE
    ld      b,(hl)
MI2CCLKHDelay:
    djnz    MI2CCLKHDelay
#endif

    out     (P_LINK),a              ; Clock pulse ends

#ifndef ZERO_WAIT_STATE
    ld      b,(hl)
MI2CCLKLDelay:
    djnz    MI2CCLKLDelay
#endif

    pop     bc

    jr      z,MI2CAck

MI2CBitCont:
    djnz    MI2CBit
#ifndef ZERO_WAIT_STATE
    pop     hl
#endif
    bit     2,c
    pop     bc
    ret     nz                      ; Return if it was a read operation
    call    MI2CStop                ; No acknowledge - send STOP
MI2CCountRetry:
    pop     hl                      ; Return address for call Start or Send

    pop     de
    ei                              ; Interrupts don't disturb anymore
    pop     hl
    pop     bc
    pop     af

    dec     e                       ; Decrement retry counter
    jp      nz,MI2CRetry            ; Retry if E>0
                                    ; Give up
    pop     de
    pop     ix
    scf                             ; CF=1: Data not sent
    ret                             ; Back to calling application

MI2CAck:
#ifndef ZERO_WAIT_STATE
    pop     hl
#endif
    dec     b                       ; Z set if B was 1 (ack bit)
    pop     bc                      ; Pushed by MI2CSend or MI2CRead
    ret     z                       ; Ack received
    jr      MI2CBackOffNoPop        ; Now it's a real error (loss of arbitration)

; Sends START pulse
; Destroys AF

MI2CStart:
    ld      a,$C0                   ; SCL=1 SDA=1
    out     (P_LINK),a
    push    bc
    ld      b,$FF                   ; To be sure that the bus isn't occupied
                                    ; (lower values increase speed...)
                                    ; $FF = about 1,3ms guard time

MI2CStartLoop:
    in      a,(P_LINK)
    cp      3
    jr      nz,MI2CBackOff          ; Bus is not free!
    djnz    MI2CStartLoop

    di                              ; Disable interrupts - and slave routines

    ld      a,$D4
    out     (P_LINK),a

#ifndef ZERO_WAIT_STATE
    ld      a,(I2CSpeed)            ; CY=0 because of cp & jr nz
    rla                             ; Same as 'sla a' here but takes less space
    ld      b,a                     ; Loop 2*(I2CSpeed) times
StartDelay:
    push    bc
    ld      b,0                     ; Every loop is 256*13 = 3328 T-cycles
StartDelay2:                        ; which takes about 0.55 ms to run on 6 MHz
    djnz    StartDelay2
    pop     bc
    djnz    StartDelay              ; With I2CSpeed 5 the delay is about 5.5 ms
                                    ; which is enough for the slave routines
                                    ; With turboed calcs, the delay is shorter,
                                    ; but the transmission is also faster with
                                    ; equal I2CSpeed settings (not tested)

#endif

    ld      a,$FC
    out     (P_LINK),a
    pop     bc
    ret

; Sends STOP pulse
; Destroys AF, B

MI2CStop:
    ld      a,$FC                   ; SCL=0 SDA=0
    out     (P_LINK),a
    ld      a,$D4                   ; SCL=1 SDA=0
    out     (P_LINK),a
#ifndef NO_SYNCH
MI2CStop2:
    in      a,(P_LINK)            ; Waitfor SCL=1 to synchronize w/ slave
    bit     1,a
    jr      z,MI2CStop2
#endif

#ifndef ZERO_WAIT_STATE
    ld      a,(I2CSpeed)
    ld      b,a
MI2CStopDelay:
    djnz    MI2CStopDelay
#endif

    ld      a,$C0                   ; SCL=1 SDA=1
    out     (P_LINK),a
    ei                              ; Interrupts don't disturb anymore
    ret

MI2CBackOff:                        ; Waits for STOP or free bus for a long time
    pop     bc
MI2CBackOffNoPop:
    ld      a,$C0
    out     (P_LINK),a
MI2CBackOff2
    ld      d,0                     ; Timeout counter, 0 -> 256
MI2CDetStop2:
    in      a,(P_LINK)
    cp      2
    jr      c,MI2CBackOff2          ; Loop if SCL=0
    jr      nz,MI2CDetStCount       ; Loop and count if SDA=1 (and SCL=1)

MI2CDetStopLoop:
    in      a,(P_LINK)
    cp      2
    jr      c,MI2CBackOff2          ; New loop if SCL=0
    jr      z,MI2CDetStopLoop       ; Loop if SDA=0

    jr      MI2CCountRetry          ; Stop detected - try again

MI2CDetStCount:
    dec     d
    jr      nz,MI2CDetStop2
    jr      MI2CCountRetry          ; Timeout - try again
